//========================================================================================
// Author: Jens Chluba (May 2003)
// Last modification: Oct 2010
// CITA, University of Toronto
// All rights reserved.
//========================================================================================
#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>
#include <cmath>

#include "Cosmos.h"
#include "recombination.Recfast.h"
#include "physical_consts.h"
#include "routines.h"
#include "Integration_routines.h"

using namespace std;

//========================================================================================
// level of communication
//========================================================================================
int Cosmos_mess_flag=0;

//========================================================================================
//
// Class Cosmos
//
//========================================================================================

//========================================================================================
// Konstructors and Destructors
//========================================================================================
void Cosmos :: init(const double h, const double T0, const double Yp, 
                    const double densities[7], 
                    const double zstart, const double zend, 
                    const int n_Xe_pts, const double Nnuval)
{
    //====================================================================================
    // The array densities[] has to contain the densities in the sequence:     
    // O_T, O_mh2, O_bh2, O_L                                                  
    // The array flag[] has to contain different flags in the sequence:        
    //====================================================================================
    h100=h; T_CMB0=T0; Y_p=Yp; Nnu=Nnuval;
    O_T=densities[0]; O_mh2=densities[1]; O_bh2=densities[2]; O_L=densities[3];
    O_T+=calc_Orel(T0, Nnu, h100);
    set_initials(Nnu);
    
    fudge=densities[4];
    if(fudge==0.0) fudge=1.14;
    
    fDM_annihilation=densities[5];
    CR_correction=(int)densities[6]; 

    //====================================================================================
    // for Xe-spline interpolations...
    //====================================================================================
    zs=zstart; ze=zend; 
    n_Xe=n_Xe_pts;

    calc_dependent_consts();
 
    splines_are_charged=0;
    
    //RF_hPlanck  =6.62606876E-34;      // Planck's constant [Js]    
    //RF_kBoltz   =1.3806503E-23;       // Boltzman's constant [J/K] 
    //RF_mElect   =9.10938188E-31;      // Electron mass [kg]  
    RF_hPlanck  =const_h*1.0e-7;      // Planck's constant [Js]    
    RF_kBoltz   =const_kB_J;          // Boltzman's constant [J/K] 
    RF_mElect   =const_me_gr*1.0e-3;  // Electron mass [kg]        
    
    return;
}

//========================================================================================
Cosmos :: Cosmos(){ spline_memory_allocated=0; }
Cosmos :: Cosmos(const double h, const double T0, const double Yp, 
                 const double densities[7], 
                 const double zstart, const double zend, 
                 const int n_Xe_pts, const double Nnu)
{ 
    spline_memory_allocated=0;
    init(h, T0, Yp, densities, zstart, zend, n_Xe_pts, Nnu); 
}

Cosmos :: ~Cosmos(){}

//========================================================================================
// private functions
//========================================================================================
void Cosmos :: set_initials(double Nnuval)
{
    g_rel=1.0+Nnuval*(7.0/8.0)*pow(4.0/11.0, 4.0/3.0);
    zsRe=1.0e+4;

    return;
}

void Cosmos :: init_splines()
{
    calc_coeff_X_spline(n_Xe, zsRe, 0.0);
    splines_are_charged=1;

    return;
}

void Cosmos :: calc_dependent_consts()
{
    h2=h100*h100; H0=100.0*h100*(1.0e+3*1.0e+2)/const_Mpc;  // H = h *100 * km/s/Mpc 
                                                            //   = h *100 * 10^5  cm/s/Mpc  
    T27=T_CMB0/2.7; Theta_CMB0=TtoTheta(T_CMB0); Theta27=T27;

    rho_c=3.0*pow(H0, 2)/(8.0*PI*const_G);                  //  1.87882e-29 h2    g cm^-3  

    rho_g=PI*PI/15.0*const_kB*T_CMB0                        //  4.47378e-34 T27^4 g cm^-3
          *pow(const_kB*T_CMB0/(const_hbar*const_cl), 3)/pow(const_cl, 2);

    rho_g_gr=rho_g/const_me_gr;
    
    O_g=rho_g/rho_c; O_gh2=O_g*h2;
    O_rel=O_g*g_rel; O_relh2=O_rel*h2;
    z_eq=O_mh2/O_relh2 -1.0;                     //  4.199628e+4/g_rel*pow(T27,-4)*O_mh2;

    O_m=O_mh2/h2; O_b=O_bh2/h2;
    O_Th2=O_T*h2; O_Lh2=O_L*h2;
    O_k=1.0-O_T; O_kh2=O_k*h2;

    //====================================================================================
    // 28.05.2008
    // we changed the definition of Nb to Nb=rho_b/mH, where mH is the neutral HI mass
    //====================================================================================
    //Nb0=O_b*rho_c/const_mp_gr/XIFAC;           //  1.12328e-5 cm^-3
    Nb0=O_b*rho_c/const_mH_gr;                   //  1.123xxe-5 cm^-3
    return;
}

//========================================================================================
int Cosmos :: recombination_history(int nzpts, double zstart, double zend, double *zarr, 
                                    double *Xe_H, double *Xe_He, double *Xe, double *TM)
{
  double *dXe=new double[nzpts];
  double *dX_H=new double[nzpts];

  int flg=recombination_history(nzpts, zstart, zend, zarr, Xe_H, Xe_He, Xe, dXe, dX_H, TM);

  delete[] dXe;
  delete[] dX_H;
  
  return flg;
}

//========================================================================================
int Cosmos :: recombination_history(int nzpts, double zstart, double zend, 
                                    double *zarr, double *Xe_H, 
                                    double *Xe_He, double *Xe, double *dXe, 
                                    double *dX_H, double *TM)
{
    double *param=new double[14];

    // set parameters for Xe_frac()
    param[0]=nzpts;
    param[1]=zstart;
    param[2]=zend;
    param[3]=Y_p;
    param[4]=T_CMB0;
    param[5]=O_m;
    param[6]=O_b;
    param[7]=O_L;
    param[8]=O_k;
    param[9]=h100;
    param[10]=Nnu;
    param[11]=fudge;
    param[12]=fDM_annihilation;
    param[13]=CR_correction;

    if(Cosmos_mess_flag>=2)
    {
        cout << "\n Cosmos :: recombination_history: " << endl;
        cout << " running recombination_history with the following parameters: " << endl;   
        
        cout << " zs: " << param[1] << "\t ze: " << param[2] 
             << "\t nzpts: " << (int)param[0] << endl;
        cout << " Y_p: " << param[3] << "\t TCMB: " << param[4] << endl;
        cout << " OmegaM: " << param[5] << "\t OmegaB: " << param[6] 
             << "\n OmegaL: " << param[7] << "\t OmegaK: " << param[8] << endl;
        cout << " Omega_gamma h2: " << O_gh2 << "\t Omega_rel h2: " << O_relh2 
             << " Nnu= " << Nnu << endl;
        cout << " Hubble constant in units of 100 km s-1 Mpc-1: " << param[9] << endl;
        cout << " Fudge factor for H recombination: " << param[11] << endl;
        cout << " DM-annihilation " << param[12] << endl;
        cout << " Chluba & Thomas correction function: " << param[13] << endl;
    }
    
    Xe_frac(param, zarr, Xe_H, Xe_He, Xe, dXe, dX_H, TM, Cosmos_mess_flag);

    delete [] param;
    return 0;
}

//========================================================================================
int Cosmos :: calc_coeff_X_spline(int M, double zstart, double zend)
{
    double *zarray=new double[M];
    double *zGSL=new double[M];
    double *Xe_H=new double[M];
    double *dXe_H=new double[M];
    double *Xe_He=new double[M];
    double *Xe=new double[M];
    double *dXe=new double[M];
    double *YGSL=new double[M];
    double *TM=new double[M];
    
    if(Cosmos_mess_flag>=1)
        cout << "\n Cosmos::calc_coeff_X_spline:" 
             << "\n calculating coeffients for X spline interpolation..." << endl;
    
    //====================================================================================
    // here arrays are filled with values
    //====================================================================================
    recombination_history(M, zsRe, zend, zarray, Xe_H, Xe_He, Xe, dXe, dXe_H, TM);
    
    //====================================================================================
    // resort zarray and Xe array so that ordering is zarray[0]<zarray[1]< ...
    //====================================================================================
    for(int j=0, i=M-1; i>=0; i--, j++) 
    {
        zGSL[j]=zarray[i];
        YGSL[j]=log(Xe[i]);
    }
    if(!spline_memory_allocated) 
        memindex_Xe=calc_spline_coeffies_JC(M, zGSL, YGSL, 
                                            "Cosmos :: calc_coeff_X_spline: Xe");
    else 
        update_spline_coeffies_JC(memindex_Xe, M, zGSL, YGSL, 
                                  "Cosmos :: calc_coeff_X_spline: Xe");

    //====================================================================================
    // dXe
    //====================================================================================
    for(int j=0, i=M-1; i>=0; i--, j++) YGSL[j]=dXe[i];
    if(!spline_memory_allocated) 
        memindex_dXe=calc_spline_coeffies_JC(M, zGSL, YGSL, 
                                             "Cosmos :: calc_coeff_X_spline: dXe");
    else 
        update_spline_coeffies_JC(memindex_dXe, M, zGSL, YGSL, 
                                  "Cosmos :: calc_coeff_X_spline: dXe");

    //====================================================================================
    // X1s
    //====================================================================================
    for(int j=0, i=M-1; i>=0; i--, j++) YGSL[j]=log(Xe_H[i]);
    if(!spline_memory_allocated) 
        memindex_XeH=calc_spline_coeffies_JC(M, zGSL, YGSL, 
                                             "Cosmos :: calc_coeff_X_spline: XeH");
    else 
        update_spline_coeffies_JC(memindex_XeH, M, zGSL, YGSL, 
                                  "Cosmos :: calc_coeff_X_spline: XeH");
                                   
    //====================================================================================
    // dXe_H
    //====================================================================================
    for(int j=0, i=M-1; i>=0; i--, j++) YGSL[j]=dXe_H[i];
    if(!spline_memory_allocated) 
        memindex_dXeH=calc_spline_coeffies_JC(M, zGSL, YGSL, 
                                              "Cosmos :: calc_coeff_X_spline: dXeH");
    else 
        update_spline_coeffies_JC(memindex_dXeH, M, zGSL, YGSL, 
                                  "Cosmos :: calc_coeff_X_spline: dXeH");
                                   
    //====================================================================================
    // rho=Te/Tg 
    //====================================================================================
    for(int j=0, i=M-1; i>=0; i--, j++) YGSL[j]=TM[i]/TCMB(zarray[i]);
    if(!spline_memory_allocated) 
        memindex_rho=calc_spline_coeffies_JC(M, zGSL, YGSL, 
                                             "Cosmos :: calc_coeff_X_spline: rho");
    else 
        update_spline_coeffies_JC(memindex_rho, M, zGSL, YGSL, 
                                  "Cosmos :: calc_coeff_X_spline: rho");
                                   
    //====================================================================================
    // XHe1s
    //====================================================================================
    for(int j=0, i=M-1; i>=0; i--, j++) YGSL[j]=log(Xe_He[i]);
    if(!spline_memory_allocated) 
        memindex_XeHe=calc_spline_coeffies_JC(M, zGSL, YGSL, 
                                              "Cosmos :: calc_coeff_X_spline: Xe_He");
    else 
        update_spline_coeffies_JC(memindex_XeHe, M, zGSL, YGSL, 
                                  "Cosmos :: calc_coeff_X_spline: Xe_He");
    
    //====================================================================================
    // clean up
    //====================================================================================
    delete[] zarray;
    delete[] zGSL;
    delete[] Xe_H;
    delete[] dXe_H;
    delete[] Xe_He;
    delete[] Xe;
    delete[] dXe;
    delete[] YGSL;
    delete[] TM;
    
    spline_memory_allocated=1;

    return 0;
}

//========================================================================================
// local spline-functions
//========================================================================================
void Cosmos :: check_splines()
{   
    if(splines_are_charged) return; 
    else
    { 
        cerr << " Cosmos::check_splines: recombination splines are not armed."
             << " Exiting. " << endl; 
        exit(0); 
    }
}

double Cosmos :: calc_spline(double z, int memindex)    
{   
    check_splines(); 
    return calc_spline_JC(z, memindex, " Cosmos::calc_spline "); 
}
double Cosmos :: calc_spline_log(double z, int memindex)
{ 
    check_splines(); 
    return exp(calc_spline_JC(z, memindex, " Cosmos::calc_spline_log ")); 
}

//========================================================================================
// public functions
//========================================================================================
void Cosmos :: display_cosmology()
{
    cout << "\n*----------------------------------------------"
         << "--------------------------------------------*" << endl;
    cout << "* Cosmos :: These are the most important cosmological"
         << " parameters used in this computation: *" << endl;
    cout << "*------------------------------------------------"
         << "------------------------------------------*\n" << endl;

    cout << " zs: " << zs << "\t\t ze: " << ze << endl;
    cout << " Y_p: " << Y_p << "\t\t fHe: " << fHe() << "\t\t TCMB: " << T_CMB0 << endl;
    cout << " OmegaM   : " << O_m   << "\t OmegaB   : " << O_b 
         << "\t OmegaL   : " << O_L   << "\t OmegaK   : " << O_k << endl;
    cout << " OmegaM h2: " << O_mh2 << "\t OmegaB h2: " << O_bh2 
         << "\t OmegaL h2: " << O_Lh2 << "\t OmegaK h2: " << O_kh2 << endl;
    cout << " Omega_gamma h2: " << O_gh2 << "\t Omega_rel h2: " << O_relh2 
         << "\t Nnu= " << Nnu << endl;
    cout << " Hubble constant in units of 100 km s-1 Mpc-1: " << h100 
         << " 1Mpc= " << const_Mpc << " cm " << endl;
    cout << " radiation-matter equality redshift z_eq: " << z_eq << endl;

    cout << "\n*----------------------------------------------"
         << "--------------------------------------------*" << endl;
    cout << "*------------------------------------------------"
         << "------------------------------------------*\n" << endl;

    return;
}

//========================================================================================
void Cosmos :: dump_cosmology(string filename)
{
    ofstream f;
    f.open(filename.c_str());
    f.precision(8);
    
    f << " zs: " << zs << "\t ze: " << ze << endl;
    f << " Y_p: " << Y_p << "\t fHe: " << fHe() << "\t TCMB: " << T_CMB0 << endl;
    f << " OmegaM   : " << O_m   << "\t OmegaB   : " << O_b 
      << "\t OmegaL   : " << O_L   << "\t OmegaK   : " << O_k << endl;
    f << " OmegaM h2: " << O_mh2 << "\t OmegaB h2: " << O_bh2 
      << "\t OmegaL h2: " << O_Lh2 << "\t OmegaK h2: " << O_kh2 << endl;
    f << " Omega_gamma h2: " << O_gh2 << "\t Omega_rel h2: " << O_relh2 
      << "\t Nnu= " << Nnu << endl;
    f << " Hubble constant in units of 100 km s-1 Mpc-1: " << h100 
      << " 1Mpc= " << const_Mpc << " cm " << endl;
    f << " radiation-matter equality redshift z_eq: " << z_eq << endl;

    f.close();

    return;
}

//========================================================================================
// Hubble-function
//========================================================================================
double Cosmos :: H(double z)
{
    //====================================================================================
    // This is equivalent with Seager 2000 ApJSS 128
    // return H0*sqrt(O_L+zp1*zp1*(O_k+zp1*O_m*(1.0+zp1/(1.0+z_eq)))); 
    // (Seager 2000 ApJSS 128)
    //====================================================================================
    double zp1=1.0+z;
    return H0*sqrt(O_L+zp1*zp1*(O_k+zp1*(O_m+O_rel*zp1)));       
}

//========================================================================================
// cosmological time
//========================================================================================
double dtdz_GSL(double logz, void *p)
{ 
    Cosmos *c =(Cosmos *) p;
    double zz=exp(logz); 
    return -zz*c->dt_dz(zz-1.0);
}

double Cosmos :: t(double z)
{
    double epsInt=1.0e-12, epsabs=0.0;
    double a=log(1.0+z), b=log(1.0+zs*1.0e+5), r=0.0;
    void *p=this;
    
    r=Integrate_gk_GSL(6, a, b, epsInt, epsabs, dtdz_GSL, p);  

    return r;
}

//========================================================================================
// Spitzer formula
//========================================================================================
double Cosmos :: t_ep(double z)                                                
{ 
    return t_eg(z)*sqrt(PI/2.0)*const_mp/(20.0*const_me)*(1.0-Y_p/2.0)
                  /(1.0-Y_p)*pow(TCMB(z)*(1.0+const_me/const_mp), 1.5); 
}

//========================================================================================
// integrals over Hubble; Integral^zh_z (1+z)^pow |dt_dz| dz
//========================================================================================
double dE_GSL(double logz, void *p)
{ 
    Cosmos *c =(Cosmos *) p;
    double z1=exp(logz); 
    return -z1*pow(z1, c->Get_power())*c->dt_dz(z1-1.0);
}

double Cosmos :: E_Int(double zh, double z, double pow)
{
    power=pow;
    double epsInt=1.0e-12, epsabs=0.0;
    double a=log(z+1.0), b=log(zh+1.0), r=0.0;
    void *p=this;

    r=Integrate_gk_GSL(6, a, b, epsInt, epsabs, dE_GSL, p);

    return r;
}

//========================================================================================
// helium nuclei
//========================================================================================
double Cosmos :: NHeII(double z)
{
  // it is assumed that NHeIII << NHeI & NHeII
  if(z<=4000) return Ne(z)-Np(z);

  // it is assumed that NHeI << NHeII & NHeIII
  return 2.0*fHe()*NH(z)+Np(z)-Ne(z);
}

double Cosmos :: NHeIII(double z)
{
  if(z<=4000) return 0.0;

  // it is assumed that NHeI << NHeII & NHeIII
  return Ne(z)-Np(z)-fHe()*NH(z);
}

double Cosmos :: NHeI(double z)
{ return fHe()*NH(z)-NHeII(z)-NHeIII(z); }

//========================================================================================
// compute Recfast system with a given initial solution at low z
//========================================================================================
int Cosmos :: recombine_using_Recfast_system(int nzpts, double zi, double ze, double fDM, 
                                             double Xe_Hi, double Xe_Hei, 
                                             double Xei, double TMi, double dXei, 
                                             const double *zarr, double *Xe_H, 
                                             double *Xe_He, double *Xe, double *TM)
{
    double *param=new double[14];

    // set parameters for Xe_frac()
    param[0]=nzpts;
    param[1]=zi;
    param[2]=ze;
    param[3]=Y_p;
    param[4]=T_CMB0;
    param[5]=O_m;
    param[6]=O_b;
    param[7]=O_L;
    param[8]=O_k;
    param[9]=h100;
    param[10]=Nnu;
    param[11]=fudge;
    param[12]=fDM;
    param[13]=0;
    
    if(Cosmos_mess_flag>=2)
    {
        cout << "\n Cosmos :: recombination_history: " << endl;
        cout << " running recombination_history with the following parameters: " << endl;   
        
        cout << " zs: " << param[1] << "\t ze: " << param[2] 
             << "\t nzpts: " << (int)param[0] << endl;
        cout << " Y_p: " << param[3] << "\t TCMB: " << param[4] << endl;
        cout << " OmegaM: " << param[5] << "\t OmegaB: " << param[6] 
             << "\n OmegaL: " << param[7] << "\t OmegaK: " << param[8] << endl;
        cout << " Omega_gamma h2: " << O_gh2 << "\t Omega_rel h2: " << O_relh2 
             << " Nnu= " << Nnu << endl;
        cout << " Hubble constant in units of 100 km s-1 Mpc-1: " << param[9] << endl;
        cout << " Fudge factor for H recombination: " << param[11] << endl;
        
        cout << "\n and initial values " << Xe_Hi << " " << Xe_Hei << " " 
             << Xei << " " << TMi << endl << endl;
    }
    
    Xe_frac_rescaled(param, zarr, Xe_H, Xe_He, Xe, TM, Xe_Hi, 
                     Xe_Hei, Xei, TMi, dXei, Cosmos_mess_flag);

    delete [] param;
    return 0;
}

//====================================================================================
// functions to compute energy release caused by acoustic damping (added 15.10.2011)
//====================================================================================
//double eff_fact_pol=8.0/9.0;
double eff_fact_pol=16.0/15.0;

double dr_s_GSL(double logz, void *p)
{ 
    Cosmos *c =(Cosmos *) p;
    double z1=exp(logz), z=z1-1.0; 
    return z1*const_cl/c->H(z)*c->c_s(z);
}

double Cosmos :: r_s(double z) // 'sound horizon'
{
    double epsInt=1.0e-7, epsabs=0.0;
    double a=log(z+1.0), b=log(1.0e+16), r=0.0;
    void *p=this;
    
    r=Integrate_using_Patterson_adaptive(a, b, epsInt, epsabs, dr_s_GSL, p);
    
    return r;
}

//====================================================================================
double dk_D_GSL(double logz, void *p)
{ 
    Cosmos *c =(Cosmos *) p;
    double z1=exp(logz), z=z1-1.0;
    double R=c->R_bg(z);
    double tau_d=c->dtau_deta(z);
    return z1*const_cl/c->H(z)/(6.0*(1.0+R))*(R*R/(1.0+R)+eff_fact_pol)/tau_d;
}

double Cosmos :: k_D(double z) // 'damping wavenumber'
{
    double epsInt=1.0e-7, epsabs=0.0;
    double a=log(z+1.0), b=log(1.0e+16), r=0.0;
    void *p=this;
    
    r=Integrate_using_Patterson_adaptive(a, b, epsInt, epsabs, dk_D_GSL, p);
    
    return sqrt(1.0/r);
}

double Cosmos :: dkm2_D_deta(double z) // eta-derivative of 1/kD^2
{
    double R=R_bg(z);
    double tau_d=dtau_deta(z);
    
    return 1.0/(6.0*(1.0+R))*(R*R/(1.0+R)+eff_fact_pol)/tau_d;
}

double Cosmos :: dkm2_D_dz(double z) // z-derivative of 1/kD^2
{
    double R=R_bg(z);
    double tau_d=dtau_deta(z);
    
    return const_cl/H(z)/(6.0*(1.0+R))*(R*R/(1.0+R)+eff_fact_pol)/tau_d;
}

//====================================================================================
// 'sound speed' of baryons
//====================================================================================
double Cosmos :: dTe_dz(double z)
{ 
    double dz=z*1.0e-6;
    return ( Te(z+dz)-Te(z-dz) )/(2.0*dz); 
}

double Cosmos :: c_sb(double z)
{
    double Xev=Xe(z);
    double mu_mol=const_mp_gr/(1.0-Y_p)/(1.0+fHe()+Xev);
    double Tb=Te(z);
    double cs2=const_kB*Tb/(const_cl*const_cl)/mu_mol;
    
//    cs2*=1.0+z/Tb/3.0*dTe_dz(z);  // Ma & Bertschinger 1995
    cs2*=1.0;                     // Lewis 2007 (10.1103/PhysRevD.76.063001), with Delta_Tg ~ 0
//    cs2*=5.0/3.0;
//    cout << " Cosmos :: c_sb " << 1.0+z/Tb/3.0*dTe_dz(z) << " " << z/Tb/3.0*dTe_dz(z) << endl;
 
    return sqrt(cs2);
}
    
double Cosmos :: k_eq()
{
    // k_eq=(2 Om H0^2 a0/aeq)^0.5
    double a0_aeq=Get_O_m()/Get_O_r()-1.0;
    
    return sqrt(2.0*Get_O_m()*a0_aeq)*Get_H0()/const_cl;
}

//====================================================================================
//====================================================================================
